/*
 wmhdaps 0.03 (2006-04-07)
 Pezhman Givy (phg@snafu.de)

 based on:
 hdaps-gl and
 
 wmcube.c 
 Version 0.98 (2000-10-20)

 Robert Kling (robkli-8@student.luth.se)
 http://boombox.campus.luth.se/projects.php

 This software is licensed through the GNU General Public Licence.
 
 See http://www.BenSinclair.com/dockapp/ for more wm dock apps.

 If you want to port wmcube to another OS the system specific code is
 sectioned the bottom of this file. See instructions there.

*/


/* Based on 
 *
 * hdapsd.c - Read from the HDAPS (HardDrive Active Protection System)
 *            and protect the drive if motion over threshold...
 *
 *            Derived from pivot.c by Robert Love.
 *
 * Copyright (C) 2005-2006 Jon Escombe <lists@dresco.co.uk>
 *                         Robert Love <rml@novell.com>
 *                         Shem Multinymous <multinymous@gmail.com>
 *
 * "Why does that kid keep dropping his laptop?"
 */


#define WMHDAPS_VERSION "0.04"
#define REV_DATE "20060501"

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>

#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <utmp.h>
#include <dirent.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>

#ifdef FREEBSD
#include <kvm.h>
#endif

#include "../wmgeneral/wmgeneral.h"
#include "../wmgeneral/misc.h"

#include "wmhdaps.xpm"
char wmcube_mask_bits[64*64];
int  wmcube_mask_width = 64;
int  wmcube_mask_height = 64;

#define CHAR_WIDTH 5
#define CHAR_HEIGHT 7
#define PI 3.1415926535

/* Magic threshold tweak factors, determined experimentally to make a
 * threshold of 10-20 behave reasonably.
 */
#define VELOC_ADJUST            30.0
#define ACCEL_ADJUST            (VELOC_ADJUST * 60)
#define AVG_VELOC_ADJUST        3.0

/* History depth for velocity average, in seconds */
#define AVG_DEPTH_SEC           0.3
#define FREQ_HZ                 50   /* sampling frequency */


//**** Graphics ***********************************

void putpixel(int x, int y, int c);
void line(int x1, int y1, int x2, int y2, int c);
void hline(int x1, int x2, int y, int c);
void triangle(int x1, int y1, int x2, int y2, int x3, int y3, int c);
void BlitString(char *name, int x, int y);
void BlitNum(int num, int x, int y);
void clearscr();
void draw();
void startup_seq();

//**** 3d specific ********************************

void setupobj(char *filename) ;
void setUpAngles();
void rotate(int xang, int yang, int zang);
int  normal(float p1[], float p2[], float p3[]);
int  luminate(float p1[], float p2[], float p3[]);
void sortz(int nofelements);

//**** Application Management, I/O etc. ***********

void print_help();
int  loadobj(char *filename);
void mem_alloc_error(void *block);
int	 scan4objects(char *dir);
int  next_object();
void die();

//**** System specific functions ******************

int  init_calc_cpu();
int  calc_cpu_total();

//**** Bar ****************************************

static int DrawBars(int percent1, int percent2, int percent3);
static void DrawLock(void);

//**** Global variables ***************************

int 	xcenter, ycenter, zoff;
double	cost[361], sint[361];
double  acost[100];
float 	**matrix;
float	**rmatrix;
int		**planes;
int		*plane_color;
int		*zorder;
int		*cline;
int		nofcoords, noflines, nofplanes;
char	*objects[1000];
int		nof_objects = 0;
int		planesORlines = 1;
char 	*pname;

float	lum_vector[3] = { 0, 0, 100 };  // Lightsource vector

#ifdef FREEBSD
static kvm_t            *kd;
static struct nlist     nlst[] = { {"_cp_time"}, {0} };
#endif

#define UPDATE_THRESHOLD	4

#if !defined(FREEBSD)
#define SYSFS_POSITION_FILE     "/sys/devices/platform/hdaps/position"
#define BUF_LEN                 32
#else
#define SYSFS_POSITION_FILE	"/dev/accel"
#define BUF_LEN			(2*sizeof(int))
#endif /* !defined(FREEBSD) */

static int rest_x;
static int rest_y;

/*
 * read_position - read the (x,y) position pair from hdaps.
 *
 * We open and close the file on every invocation, which is lame but due to
 * several features of sysfs files:
 *
 *	(a) Sysfs files are seekable.
 *	(b) Seeking to zero and then rereading does not seem to work.
 *
 * If I were king--and I will be one day--I would of made sysfs files
 * nonseekable and only able to return full-size reads.
 */
static int read_position (int *x, int *y)
{
#if !defined(FREEBSD)
	char buf[BUF_LEN];
#else
	int buf[2];
#endif /* !defined(FREEBSD) */

	int fd, ret;

//read_again:
	memset(buf, 0, sizeof(buf));

	fd = open (SYSFS_POSITION_FILE, O_RDONLY);
	if (fd < 0) {
		perror ("open");
		return fd;
	}	

	ret = read (fd, buf, BUF_LEN);
	if (ret < 0) {
		perror ("read");
		goto out;
	} else if (ret == 0) {
		fprintf (stderr, "error: unexpectedly read zero!\n");
		ret = 1;
		goto out;
	}
	ret = 0;

#if !defined(FREEBSD)
	if (sscanf (buf, "(%d,%d)\n", x, y) != 2)	{
		fprintf( stderr, "sscanf() error!\n" );
		ret = 1;
	}
#else
	*x = buf[0];
	*y = buf[1];
#endif /* !defined(FREEBSD) */

out:
	if (close (fd))
		perror ("close");

	//if(ret == 0 && *x==0 && *y==0)	goto read_again;

	return ret;
}

static int check_for_update (int *val_x, int *val_y)
{
	int ret, x, y, do_update = 0;
	static int my_val_x=0, my_val_y=0;
	static int firsttime = 1;
	static int counter = 0;

	ret = read_position (&x, &y);
	if (ret)
		return -1;

	x -= rest_x;
	y -= rest_y;

	/* only update if we surpass our threshold, to minimize jitter ... */ 


	if (firsttime || abs(abs (x) - abs(my_val_x)) > UPDATE_THRESHOLD || abs(abs (y) - abs(my_val_y)) > UPDATE_THRESHOLD)	{
		*val_x = x;
		*val_y = y;
		my_val_x = x;
		my_val_y = y;
		do_update = 1;
		counter = 0;
	}
	
	if(firsttime)	{
		firsttime = 0;
		do_update = 1;
	}
			
	/*
	if(do_update)	{
		fprintf(stderr, "position=(%d,%d)\n", *val_x, *val_y);
		fprintf(stderr, "--------------\n");
	}
	*/

	return do_update;
}

int check_protect(char *device)
{
	char buf[10];
	char filename[256];
	int ret, fd;
	struct stat statbuf;


	memset(buf, 0, sizeof(buf));

	sprintf(filename, "/sys/block/%s/queue/protect", device);
	if(stat(filename, &statbuf))	return 0;

	fd = open (filename, O_RDONLY);
	if (fd < 0) {
		perror ("open");
		return fd;
	}	

	ret = read (fd, buf, 1);
	if (ret < 0) {
		perror ("read");
		goto out;
	} else if (ret == 0) {
		fprintf (stderr, "error: unexpectedly read zero!\n");
		ret = 1;
		goto out;
	}

	ret = atoi(buf);

out:
	if (close (fd))
		perror ("close");

	return ret;
}
	

int analyze(int x, int y, double unow, double threshold, int *veloc_percent, int *accel_percent, int *avg_veloc_percent) {
	static int x_last = 0, y_last = 0;
	static double unow_last = 0, x_veloc_last = 0, y_veloc_last = 0;
	static double x_avg_veloc = 0, y_avg_veloc = 0;
	static int history = 0; /* how many recent valid samples? */

	double udelta, x_delta, y_delta, x_veloc, y_veloc, x_accel, y_accel;
	double veloc_sqr, accel_sqr, avg_veloc_sqr;
	double exp_weight;
	char reason[4];
	int ret = 0;
	int verbose = 0;
	
	/* compute deltas */
	udelta = unow - unow_last;
	x_delta = x - x_last;
	y_delta = y - y_last;

	/* compute velocity */
	x_veloc = x_delta/udelta;
	y_veloc = y_delta/udelta;
	veloc_sqr = x_veloc*x_veloc + y_veloc*y_veloc;

	/* compute acceleration */
	x_accel = (x_veloc - x_veloc_last)/udelta;
	y_accel = (y_veloc - y_veloc_last)/udelta;
	accel_sqr = x_accel*x_accel + y_accel*y_accel;

	/* compute exponentially-decaying velocity average */
	exp_weight = udelta/AVG_DEPTH_SEC; /* weight of this sample */
	exp_weight = exp_weight/(1+exp_weight); /* softly clamped to 1 */
	x_avg_veloc = exp_weight*x_veloc + (1-exp_weight)*x_avg_veloc;
	y_avg_veloc = exp_weight*y_veloc + (1-exp_weight)*y_avg_veloc;
	avg_veloc_sqr = x_avg_veloc*x_avg_veloc + y_avg_veloc*y_avg_veloc;

	*veloc_percent = (int)(veloc_sqr/(threshold*threshold * VELOC_ADJUST*VELOC_ADJUST)*100.0);
	*accel_percent = (int)(accel_sqr/(threshold*threshold * ACCEL_ADJUST*ACCEL_ADJUST)*100.0);
	*avg_veloc_percent = (int)(avg_veloc_sqr/(threshold*threshold * AVG_VELOC_ADJUST*AVG_VELOC_ADJUST)*100.0);
	if(*veloc_percent > 100)	*veloc_percent=100;
	if(*accel_percent > 100)	*accel_percent=100;
	if(*avg_veloc_percent > 100)	*avg_veloc_percent=100;

	//fprintf(stderr, "veloc_percent=%d, accel_percent=%d\n", *veloc_percent, *accel_percent);

	/* Threshold test (uses Pythagoras's theorem) */
	strcpy(reason, "   ");
	if (veloc_sqr > threshold*threshold * VELOC_ADJUST*VELOC_ADJUST ) {
		ret = 1;
		reason[0]='V';
	}
	if (accel_sqr > threshold*threshold * ACCEL_ADJUST*ACCEL_ADJUST ) {
		ret = 1;
		reason[1]='A';
	}
	if (avg_veloc_sqr > 
	    threshold*threshold * AVG_VELOC_ADJUST*AVG_VELOC_ADJUST ) {
		ret = 1;
		reason[2]='X';
	}

	if (verbose) {
		printf("dt=%5.3f  "
		       "delta=(%3g,%3g)  "
		       "veloc=(%6.1f,%6.1f)*%g  " /* For easy comparison to threshold. */
		       "accel=(%6.1f,%6.1f)*%g  " 
		       "avg_veloc=(%6.1f,%6.1f)*%g  "
		       "%s\n",
		       udelta,
		       x_delta, y_delta,
		       x_veloc/VELOC_ADJUST,
		       y_veloc/VELOC_ADJUST,
		       VELOC_ADJUST*1.0,
		       x_accel/ACCEL_ADJUST,
		       y_accel/ACCEL_ADJUST,
		       ACCEL_ADJUST*1.0,
		       x_avg_veloc/AVG_VELOC_ADJUST,
		       y_avg_veloc/AVG_VELOC_ADJUST,
		       AVG_VELOC_ADJUST*1.0,
		       reason);
	}

	if (udelta>1.0) { /* Too much time since last (resume from suspend?) */
		history = 0;
		x_avg_veloc = y_avg_veloc = 0;
	}

	if (history<2) { /* Not enough data for meaningful result */
		ret = 0;
		++history;
	}

	x_last = x;
	y_last = y;
	x_veloc_last = x_veloc;
	y_veloc_last = y_veloc;
	unow_last = unow;

	return ret;
}

double get_utime (void)
{
	struct timeval tv;
	int ret = gettimeofday(&tv, NULL);
	if (ret) {
		perror("gettimeofday");
		exit(1);
	}
	return tv.tv_sec + tv.tv_usec/1000000.0;
}

int main(int argc, char **argv)
{
	int		j;
	//long	screen_speed = 10000;    // microseconds between screen updates (approx.)
	int		but_stat = -1;
	XEvent	Event;

	char	*obj_filename = {""};
	int 	cube_color = 1;
	int 	c = 0;
	double treshold = 15.0;
#if !defined(FREEBSD)
	char device[256] = "hda";
#else
	char device[256] = "ad0";
#endif /* !defined(FREEBSD) */
	int x = 0, y = 0;

	pname = strrchr(argv[0], '/');
	if (pname == NULL) pname = argv[0];

	srand((unsigned)time(NULL));
	opterr = 0;
	
	while ((c = getopt (argc, argv, "ho:d:")) != -1) {
		switch (c)
		{
			case 'h':
				print_help();
				return 1;
			case 'o':
				obj_filename = optarg;
				break;
			case 'd':
				strcpy(device, optarg);
				break;
			case 's':
				treshold = (double)atof(optarg);
				break;
			case '?':
					print_help();
					return 1;
			default:
				abort();
		}
	}

	/*
	 * Scan directory for .wmc files and choose one randomly. If the user 
	 * specified a particular file, load that one.
	 */

#ifndef SOLARIS  // scan4objects doesnt work on Solaris, load object immediatly
	scan4objects(obj_filename);
	
	if (nof_objects != 0)
		next_object();
	else
#endif
		setupobj(obj_filename);
	 	
	/*
	 * Various initializion stuff for the 3d-engine etc.
	 */

	setUpAngles();

	createXBMfromXPM(wmcube_mask_bits, wmhdaps_xpm, wmcube_mask_width, wmcube_mask_height);
	openXwindow(argc, argv, wmhdaps_xpm, wmcube_mask_bits, wmcube_mask_width, wmcube_mask_height);

	startup_seq();
	
	// index, left, top, right, bottom
	AddMouseRegion(1,  45,  45, 58, 58);  // + Zoom In
	AddMouseRegion(5,  5, 45, 20, 58);	  // - Zoom Out
	//AddMouseRegion(3,  21, 45, 45, 58);	  //   Show cpu-load
 	AddMouseRegion(2,  5, 5, 55, 45);	  // Everywhere else (almost) to change object

 	/*
 	 * Main loop begins here
 	 */

	{
		int ret = read_position (&rest_x, &rest_y);
		if(ret)	return 1;
	}

	while (1)
	{
		double unow = get_utime(); /* microsec */
		int ret, call_again=1, protected;
		static int veloc_percent=0, accel_percent=0, avg_veloc_percent=0;

		usleep (1000000/FREQ_HZ);

		ret = check_for_update(&x, &y);
		if(ret < 0)	exit(-1);
		
		protected = check_protect(device);
			
		if(ret > 0 || call_again || protected)	{
			int x_rot, y_rot;
			
			x_rot = x<0? (360+x/2):x/2;
			y_rot = y<0? (360+y/2):y/2;
			rotate(360-y_rot, 0, 360-x_rot);

			clearscr();
			analyze(x, y, unow, treshold, &veloc_percent, &accel_percent, &avg_veloc_percent);
			draw(cube_color);
			call_again = DrawBars(avg_veloc_percent, veloc_percent, accel_percent);
			if(protected)	DrawLock();
			RedrawWindow();
		}

		
      // X Events
		
		while (XPending(display))
		{
			XNextEvent(display, &Event);
			switch (Event.type)
			{
				case Expose:
					RedrawWindow();
					break;
				case DestroyNotify:
					XCloseDisplay(display);
					exit(0);
					break;
				case ButtonPress:
					j = CheckMouseRegion(Event.xbutton.x, Event.xbutton.y);
					but_stat = j;
					
					break;
				case ButtonRelease:
					j = CheckMouseRegion(Event.xbutton.x, Event.xbutton.y);
					
					switch(j)
					{
						case 1:
							if (zoff > 750) {
								clearscr();
								BlitString("",48,48);
								rotate(0, 0, 0);
								draw(cube_color);
								RedrawWindow();
								zoff -= 150;
							}
							break;

						case 2:
							next_object();
							break;

							/*
						case 3:
							if (show_load == 1) show_load = 0; else show_load = 1;						
							ycenter = 15 - 2*show_load;
							break;
							*/

						case 5:
							clearscr();
							BlitString("",11,49);
							rotate(0, 0, 0);
							draw(cube_color);
							RedrawWindow();
							zoff += 150;
							break;
					}
			}	
			break;
		}
		//usleep(screen_speed);
	}

	/*
	 * Free up memory used by the object (dirty...)
	 */

	free(matrix);
	free(rmatrix);
	free(cline);	

	return 1;
}


//**** Graphics ***********************************
//*************************************************

void startup_seq() 
{
	char tmp[32];
	//int oldzoff = 3600;

	sprintf(tmp,"V%s",WMHDAPS_VERSION);
	
	RedrawWindow();
	BlitString("WMHDAPS",13,22);
	BlitString(tmp,15,31);
	RedrawWindow();
	RedrawWindow();
	usleep(3000000);
	RedrawWindow();
	
	/*
	zoff = 1200;

	for (;zoff < oldzoff; zoff += 35)
	{
		rotate((zoff-1200)/8,0,0);
		clearscr();
		BlitString("WMCUBE",13,22);
		BlitString(tmp,15,31);
		draw(1);
		RedrawWindow();
		usleep(9000);
	}
	*/

	//zoff = 3600;
	zoff = 3600-3600/4;
}	

void draw(int color)
{
	int i;

	if (planesORlines) {

		sortz(nofplanes);
		for (i = 0; i < nofplanes; i++) {
	    	if (normal(rmatrix[planes[zorder[i]][0]], rmatrix[planes[zorder[i]][1]], rmatrix[planes[zorder[i]][2]]) > 0) {	
			
			triangle(xcenter+rmatrix[planes[zorder[i]][0]][0], ycenter+rmatrix[planes[zorder[i]][0]][1],
				 	 xcenter+rmatrix[planes[zorder[i]][1]][0], ycenter+rmatrix[planes[zorder[i]][1]][1],
					 xcenter+rmatrix[planes[zorder[i]][2]][0], ycenter+rmatrix[planes[zorder[i]][2]][1], plane_color[zorder[i]]);
			}
		}
		
	} else {
		for (i = 0; i < noflines; i += 2)
			line(xcenter+rmatrix[cline[i  ]-1][0], ycenter+rmatrix[cline[i  ]-1][1],
		 	 	 xcenter+rmatrix[cline[i+1]-1][0], ycenter+rmatrix[cline[i+1]-1][1],color);
	}
}

void putpixel(int x, int y,int c)
{
	if ((x > 4) && (x < 59) && (y > 4) && (y < 59))
		copyXPMArea(160-c,0,1,1,x,y);
}

void hline(int x1, int x2, int y, int c)
{
	if ((y > 4) && (y < 59)) {
		if (x1 <= 4) x1 = 5;  else if (x1 > 57) return; 
		if (x2 > 57) x2 = 57; else if (x2 <= 4) return; 

		copyXPMArea(105, 56+c + 9*(c/18), x2-x1, 1, x1, y);
	}
}

void triangle(int x1, int y1, int x2, int y2, int x3, int y3, int c) // Draws a filled triangle
{
    int k,k2,x,x_2,i, tmp1; 

    int x1t, x2t;
			    
    if (y3<y2)
    {
		tmp1=y2;
		y2=y3;
		y3=tmp1;
		tmp1=x2;
		x2=x3;
		x3=tmp1;
    }

    if (y2<y1)
    {
		tmp1=y1;
		y1=y2;
		y2=tmp1;
		tmp1=x1;
		x1=x2;
		x2=tmp1;
    }

    if (y3<y2)
    {
		tmp1=y2;
		y2=y3;
		y3=tmp1;
		tmp1=x2;
		x2=x3;
		x3=tmp1;
    }
    
    if (y1!=y3)	k=((x1-x3) << 6) / (y1-y3);
    else k=(x1-x3) << 6;

	if (y1!=y2)	k2=((x1-x2) << 6) / (y1-y2);
    else k2=(x1-x2) << 6;

    x=x1 << 6;
    x_2=x;
    i=y1;

    if (i!=y2)
	do
	{
	    x+=k;
	    x_2+=k2;
	    i++;
	  
		if ((x1t = x >> 6) >  (x2t = x_2 >> 6))
			hline(x2t, x1t, i, c);
		else
			hline(x1t, x2t, i, c);
	}
	while (i!=y2);

    if (i==y3) return;

    if (y2!=y3) k2=((x2-x3) << 6) / (y2-y3);
    else k2=((x2-x3) << 6);

    x_2=x2 << 6;
    i=y2;
    do
    {
		x+=k;
		x_2+=k2;
		i++;

		if ((x1t = x >> 6) >  (x2t = x_2 >> 6))
			hline(x2t, x1t, i, c);
		else
			hline(x1t, x2t, i, c);
    }
    while (i!=y3);
}

void clearscr()
{
	copyXPMArea(78,0,56,56,4,4);
}

// Blits a string at given co-ordinates
void BlitString(char *name, int x, int y)
{
	int		i;
	int		c;
	int		k;
	
	k = x;
	
	copyXPMArea(73,64,1,8,k-1,y);
	
	for (i=0; name[i]; i++)
	{  
		c = toupper(name[i]); 
      //printf("%c",c);
		
		if (c >= 'A' && c <= 'Z')
		{   // its a letter
			c -= 'A';
			if ( k > -2) copyXPMArea(c * 6, 74, 6, 8, k, y);
			k += 6;
		} else
		if (c >= '0' && c<= ':')
		{   // its a number or symbol
			c -= '0';
			if ( k > -2) copyXPMArea(c * 6, 64, 6, 8, k, y);
			k += 6;
		} else
		if (c == 246) {
			if ( k > -2) copyXPMArea(0, 84, 6, 9, k, y);
			k += 6;
		} else
		if (c == 228) {
			if ( k > -2) copyXPMArea(6, 84, 6, 9, k, y);
			k += 6;
		} else
		if (c == 229) {
			if ( k > -2) copyXPMArea(12, 84, 6, 9, k, y);
			k += 6;
		} else
		{   // its a blank or something else
			if ( k > -2) copyXPMArea(73,64,6,8,k,y);
			k += 6;
		}
		if (k >= 58) break;
	}
	copyXPMArea(73,64,1,8,k,y);
	
}

static int DrawBars(int percent1, int percent2, int percent3)
{
	static int max1_counter=0, max2_counter=0, max3_counter=0;
	static int max1=0, max2=0, max3=0;
	int segment_count=7;
	int segment_width=5;
	int on_w;

	on_w = segment_count*percent1/100;
	copyXPMArea(32,84,on_w*segment_width,3,4+16,6);
	if(max1 < percent1)	{
		max1 = percent1;
		max1_counter = 50;
	}

	on_w = segment_count*percent2/100;
	copyXPMArea(32,84,on_w*segment_width,3,4+16,6+3);
	if(max2 < percent2)	{
		max2 = percent2;
		max2_counter = 50;
	}
	
	on_w = segment_count*percent3/100;
	copyXPMArea(32,84,on_w*segment_width,3,4+16,6+6);
	if(max3 < percent3)	{
		max3 = percent3;
		max3_counter = 50;
	}

	on_w = (segment_count*max1/100)*segment_width-segment_width;
	if(on_w >= 0 && max1 && max1_counter-->0)	
		copyXPMArea(32+on_w, 84, segment_width, 3, 4+16+on_w, 6);
	else	
		max1=0;
	
	on_w = (segment_count*max2/100)*segment_width-segment_width;
	if(on_w >= 0 && max2 && max2_counter-->0)	
		copyXPMArea(32+on_w, 84, segment_width, 3, 4+16+on_w, 6+3);
	else	
		max2=0;
	
	on_w = (segment_count*max3/100)*segment_width-segment_width;
	if(on_w >= 0 && max3 && max3_counter-->0)
		copyXPMArea(32+on_w, 84, segment_width, 3, 4+16+on_w, 6+6);
	else	
		max3=0;


	return max1 || max2 || max3;
}

static void DrawLock(void)
{
	copyXPMArea(79,62,7,9,7,6);
}

// Blits number to given coordinates.. two 0's, right justified

void BlitNum(int num, int x, int y)
{
	char buf[1024];
	int newx=x;
	
	if (num > 99)
	{
		newx -= CHAR_WIDTH;
	}
	
	if (num > 999)
	{
		newx -= CHAR_WIDTH;
	}
	
	sprintf(buf, "%02i", num);
	
	BlitString(buf, newx, y);
}

void line(int x1, int y1, int x2, int y2, int c)
{
	int i, deltax, deltay, numpixels,
	d, dinc1, dinc2,
	x, xinc1, xinc2,
	y, yinc1, yinc2;
	
	deltax = abs(x2 - x1);
	deltay = abs(y2 - y1);
	if (deltax >= deltay)
	{
		numpixels = deltax + 1;
		d = (deltay << 1) - deltax;
		dinc1 = deltay << 1;
		dinc2 = (deltay - deltax) << 1;
		xinc1 = 1;
		xinc2 = 1;
		yinc1 = 0;
		yinc2 = 1;
	}
	else
	{
		numpixels = deltay + 1;
		d = (deltax << 1) - deltay;
		dinc1 = deltax << 1;
		dinc2 = (deltax - deltay) << 1;
		xinc1 = 0;
		xinc2 = 1;
		yinc1 = 1;
		yinc2 = 1;
	}
	if (x1 > x2)
	{
		xinc1 = - xinc1;
		xinc2 = - xinc2;
	}
	if (y1 > y2)
	{
		yinc1 = - yinc1;
		yinc2 = - yinc2;
	}
	x = x1;
	y = y1;
	for (i=1; i<numpixels; i++)
	{
		putpixel(x, y, c);
		if (d < 0)
		{
			d = d + dinc1;
			x = x + xinc1;
			y = y + yinc1;
		}
		else
		{
			d = d + dinc2;
			x = x + xinc2;
			y = y + yinc2;
		}
	}
}


//**** 3d specific ********************************
//*************************************************

void rotate(int xang, int yang, int zang)
{
	float tx, ty, tz;
	int i;
	
	for (i = 0; i < nofcoords; i++) 
	{
		tx = cost[yang]*matrix[i][0]-sint[yang]*matrix[i][2];
		tz = sint[yang]*matrix[i][0]+cost[yang]*matrix[i][2];
		ty = cost[zang]*matrix[i][1]-sint[zang]*tx;

		rmatrix[i][0] = (cost[zang]*tx+sint[zang]*matrix[i][1]);
		rmatrix[i][1] = (sint[xang]*tz+cost[xang]*ty);
		rmatrix[i][2] = (cost[xang]*tz-sint[xang]*ty);
	}

	/*
	if (planesORlines)
		for (i = 0; i < nofplanes; i++)
			if (normal(rmatrix[planes[i][0]], rmatrix[planes[i][1]], rmatrix[planes[i][2]]) > 0)
				plane_color[i] = luminate(rmatrix[planes[i][0]], rmatrix[planes[i][1]], rmatrix[planes[i][2]]);
	*/

	for (i = 0; i < nofcoords; i++) {	
  		// Perspective correcting lines...
		rmatrix[i][0] = (rmatrix[i][0] *256) / (2*rmatrix[i][2] - zoff) + xcenter;
		rmatrix[i][1] = (rmatrix[i][1] *256) / (2*rmatrix[i][2] - zoff) + ycenter;
	}
}

void sortz(int nofelements) {   // Insertion-sort the planes in increasing z-distance

	int i, j, k;
	float key;
	float temparr[nofelements];

	for (i = 0; i < nofelements; i++) 
    {
   		zorder[i] = i;
   		temparr[i] = rmatrix[planes[i][0]][2]+rmatrix[planes[i][1]][2]+rmatrix[planes[i][2]][2];
    }

	for (j = 1; j < nofelements; j++) {
		
		key = temparr[j];
		k = zorder[j];
		i = j - 1;
			
		while ((i > -1) && (temparr[i] > key)) {
			int k = i+1;
			temparr[k] = temparr[i];
			zorder[k] = zorder[i--];
		}

		zorder[i+1] = k;										
		temparr[i+1] = key;
	}
}

int normal(float p1[], float p2[], float p3[])
{
		return ((p1[0]-p3[0])*(p2[1]-p3[1])-(p2[0]-p3[0])*(p1[1]-p3[1]));
}

int luminate(float p1[], float p2[], float p3[])
{
      double x1 = (float)(p1[0]-p3[0]), y1 = (float)(p1[1]-p3[1]),  z1 = (float)(p1[2]-p3[2]);
      double x2 = (float)(p2[0]-p3[0]), y2 = (float)(p2[1]-p3[1]),  z2 = (float)(p2[2]-p3[2]);
      double nx = y1*z2-y2*z1, ny =-(x1*z2-x2*z1),nz = x1*y2-y1*x2;

	  return (int)(53 * (acost[(int)(50 + 50*(nx*lum_vector[0]+ny*lum_vector[1]+nz*lum_vector[2])/
				  					(sqrt(nx*nx+ny*ny+nz*nz)*
				   					sqrt(lum_vector[0]*lum_vector[0]+lum_vector[1]*lum_vector[1]+
				   					lum_vector[2]*lum_vector[2])))] / PI));	

	  // Do I smell optimization? :-)				   						
}

void setUpAngles()
{
	int i;
	for (i = 0; i < 361; i++) {
		cost[i] = cos((double)i*(2*PI/(double)360));
		sint[i] = sin((double)i*(2*PI/(double)360));
	}

	for (i = 0; i < 100; i++) acost[i] = acos((double)(-50+i)/50);
}

void setupobj(char *filename) 
{
	int i, j = 0;
	int biggest = 0;
	float scale = 1;

	xcenter = 16;
	ycenter = 15;
	
	if (strcmp(filename,"") != 0)
		loadobj(filename);
	else
	{
		nofcoords = 8;
		noflines = 0;
		nofplanes = 12;
		planesORlines = 1;

		matrix = 		(float **)malloc(nofcoords*sizeof(float *)); mem_alloc_error(matrix);
		planes = 		(int **)malloc(nofplanes*sizeof(int *)); mem_alloc_error(planes);
		plane_color =	(int *)malloc(nofplanes*sizeof(int)); mem_alloc_error(plane_color);
		zorder = 		(int *)malloc(nofplanes*sizeof(int)); mem_alloc_error(zorder);

		for (i = 0; i < nofplanes; i++) zorder[i] = i;
		
		for (i = 0; i < nofcoords; i++) { 
			matrix[i] = (float *)malloc(3*sizeof(float));
			mem_alloc_error(matrix[i]);
		}

		for (i = 0; i < nofplanes; i++) { 
			planes[i] = (int *)malloc(3*sizeof(int));
			mem_alloc_error(planes[i]);
		}

		cline = (int *)malloc((noflines+1)*sizeof(int)); mem_alloc_error(cline);
	
		matrix[0][0] = 180; matrix[0][1] = -180; matrix[0][2] =  -180; // 0
		matrix[1][0] =  180; matrix[1][1] = -180; matrix[1][2] =  180; // 1
		matrix[2][0] =  -180; matrix[2][1] =  -180; matrix[2][2] =  180; // 2
		matrix[3][0] = -180; matrix[3][1] =  -180; matrix[3][2] =  -180; // 3
		matrix[4][0] = 180; matrix[4][1] = 100; matrix[4][2] = -180; // 4
		matrix[5][0] =  180; matrix[5][1] = -150; matrix[5][2] = -180; // 5
		matrix[6][0] =  -180; matrix[6][1] =  -150; matrix[6][2] = -180; // 6
		matrix[7][0] = -180; matrix[7][1] =  100; matrix[7][2] = -180; // 7

		planes[0][0] = 0; planes[0][1] = 1; planes[0][2] = 2;
		planes[1][0] = 2; planes[1][1] = 3; planes[1][2] = 0;		
		planes[2][0] = 1; planes[2][1] = 2; planes[2][2] = 3;
		planes[3][0] = 3; planes[3][1] = 0; planes[3][2] = 1;

		planes[4][0] = 2; planes[4][1] = 1; planes[4][2] = 0;
		planes[5][0] = 0; planes[5][1] = 3; planes[5][2] = 2;		
		planes[6][0] = 1; planes[6][1] = 0; planes[6][2] = 3;
		planes[7][0] = 3; planes[7][1] = 2; planes[7][2] = 1;

		planes[8][0] = 4;  planes[8][1] = 5;  planes[8][2] = 6;
		planes[9][0] = 6;  planes[9][1] = 7;  planes[9][2] = 4;		
		planes[10][0] = 5; planes[10][1] = 4; planes[10][2] = 7;
		planes[11][0] = 7; planes[11][1] = 6; planes[11][2] = 5;
	}

 	rmatrix = (float **)realloc(rmatrix,nofcoords*sizeof(float *)); mem_alloc_error(rmatrix);
 	for (i = 0; i < nofcoords; i++) { 
 		rmatrix[i] = (float *)malloc(3*sizeof(float));
 		mem_alloc_error(rmatrix[i]);
 	}

	/*
	 * Find the longest discance between all coordinates relative to the origin
	 */

	for (i = 0; i < nofcoords; i++) {
		j = (int)sqrt((pow(matrix[i][0],2)+pow(matrix[i][1],2)+pow(matrix[i][2],2)));
		if (j > biggest) biggest = j;
	}

	/*
	 * Scale every coordinate using the calculated factor
	 */

	scale = 280 / (float)biggest;	

	for (i = 0; i < nofcoords; i++) {
		matrix[i][0] *= scale;
		matrix[i][1] *= scale;
		matrix[i][2] *= scale;
	}
}

//**** Application Management, I/O etc. ***********
//*************************************************

void print_help() {
	printf("\nwmhdaps %s (%s)\n\n", WMHDAPS_VERSION, REV_DATE);
	
#ifndef SOLARIS
	printf("  -o <filename|dir>: load external 3d-object(s).\n\n");
#else
	printf("  -o <filename>    : load external 3d-object.\n\n");
#endif

	printf("  -d <device>      : device to be observed for protection (default: hda)\n");
	printf("  -s <sensitivity> : sensitivity to operate with (default: 15)\n");
	printf("  -h               : display this helptext.\n\n");	
}

void die()
{
	fprintf(stderr, "%s: exiting", pname);
	exit (1);
}

#ifndef SOLARIS 	// scan4objects doesnt work on Solaris because of alphasort
int scan4objects(char *dir) 
{
	struct dirent **names;
	int n;

	n = scandir(dir,&names,0,alphasort);
		
	while (n-- > 0) 
		if (strstr(names[n]->d_name,".wmc") != NULL)
		{			
			objects[nof_objects] = (char *)malloc(strlen(dir)+strlen(names[n]->d_name)+2);
			strcpy(objects[nof_objects],dir);
			if (dir[strlen(dir)] != '/') strcat(objects[nof_objects],"/");
			strcat(objects[nof_objects++],names[n]->d_name);
		}

	return nof_objects;
}
#endif

int next_object()
{
	if (nof_objects == 0) return -1;
	setupobj(objects[rand() % (nof_objects )]);
	
	return 0;
}

int loadobj(char *filename) {
	
	FILE *fp;
	char tmp[64] = {""};
	int i = 0, counter = 1;

	//printf("\nLoading file %s...",filename); fflush(stdout);
	
	if ((fp = fopen(filename,"rt")) == NULL) {
		printf("\nERROR: wmCube object-file not found (%s).\n\n",filename);
		exit(0);
	}

	if (fscanf(fp,"%s",tmp) != 1) {
		printf("\nError in objectfile: %s\n\n", strerror(errno));
		fclose(fp);
		exit(0);
	}
	
	if (strcmp(tmp,"WMCUBE_COORDINATES") != 0) { 
		printf("\nError in objectfile: it must start with WMCUBE_COORDINATES\n\n");
		fclose(fp);
		exit(0);
	}

	if (fscanf(fp,"%s",tmp) != 1) {
		printf("\nError in objectfile: %s\n\n", strerror(errno));
		fclose(fp);
		exit(0);
	}
	counter = atoi(tmp);

	while ((strcmp(tmp,"WMCUBE_LINES") != 0) && (strcmp(tmp,"WMCUBE_PLANES") != 0)) {		

	 	matrix = (float **)realloc(matrix,(i+1)*sizeof(float *)); mem_alloc_error(matrix);
		matrix[i] = (float *)malloc(3*sizeof(float)); mem_alloc_error(matrix[i]);
		if (fscanf(fp,"%f %f %f",&matrix[i][0],&matrix[i][1],&matrix[i][2]) != 3) {
			printf("\nError in objectfile: %s\n\n", strerror(errno));
			fclose(fp);
			exit(0);
		}
		//printf("\n%d: %f %f %f",atoi(tmp), matrix[i][0],matrix[i][1],matrix[i][2]);

		if (atoi(tmp) != (++i)) { 
			
			printf("\nError in objectfile (WMCUBE_COORDINATES section):\n"
					 "the coordinates must be listed in order 1..n\n\n");
			fclose(fp);
			exit(0);
		}
		if (fscanf(fp,"%s",tmp) != 1) {
			printf("\nError in objectfile: %s\n\n", strerror(errno));
			fclose(fp);
			exit(0);
		}

		if (feof(fp)) {
			printf("\nError in objectfile: you must have a section WMCUBE_LINES or WMCUBE_PLANES\n\n");
			fclose(fp);
			exit(0);
		}
	}
 
	nofcoords = i;
	i = 0;

	if (strcmp(tmp,"WMCUBE_LINES") == 0) {

		planesORlines = 0;
		while (1) {
		
			cline = (int *)realloc(cline,(i+2)*sizeof(int)); mem_alloc_error(cline);
			if (fscanf(fp,"%d %d",&cline[i],&cline[i+1]) != 2) {
				printf("\nError in objectfile: %s\n\n", strerror(errno));
				fclose(fp);
				exit(0);
			}
			i+=2;
			//printf("\n%d %d",cline[i-2],cline[i-1]);
			if (feof(fp)) break;
		
			if (cline[i-2] > nofcoords || cline[i-1] > nofcoords) { 
				printf("\nError in objectfile (WMCUBE_LINES section):\n"
					   "coordinates %d or/and %d doesnt exist\n\n",cline[i-2],cline[i-1]);
				fclose(fp);
				exit(0);
			}
		}
		noflines = i-2;
	}
	else if (strcmp(tmp,"WMCUBE_PLANES") == 0) {
	
		planesORlines = 1;
		while (1) {	
			planes = (int **)realloc(planes,(i+1)*sizeof(int *)); mem_alloc_error(planes);
			planes[i] = (int *)malloc(3*sizeof(int)); mem_alloc_error(planes[i]);
			if (fscanf(fp,"%d %d %d",&planes[i][0],&planes[i][1],&planes[i][2]) != 3) {
				printf("\nError in objectfile: %s\n\n", strerror(errno));
				fclose(fp);
				exit(0);
			}
			//printf("\n%d: %d %d %d",i,planes[i][0],planes[i][1],planes[i][2]);

			planes[i][0]--; planes[i][1]--; planes[i][2]--;
			//printf("\n%d: %d %d %d\n",i,planes[i][0],planes[i][1],planes[i][2]);
			
			if (feof(fp)) break;
		
			if (planes[i][0] > nofcoords || planes[i][1] > nofcoords || planes[i][2] > nofcoords) { 
				printf("\nError in objectfile (WMCUBE_PLANES section):\n"
				 	   "coordinates %d or/and %d or/and %d doesnt exist\n\n",planes[i][0],planes[i][1],planes[i][2]);
				fclose(fp);
				exit(0);
			}
			i++;		
		}
		nofplanes = i;
		plane_color =	(int *)malloc(nofplanes*sizeof(int)); mem_alloc_error(plane_color);
		zorder =		(int *)malloc(nofplanes*sizeof(int)); mem_alloc_error(zorder);
		for (i = 0; i < nofplanes; i++) zorder[i] = i;
		
	} else {
		printf("\nError in objectfile: you must have a section WMCUBE_LINES or WMCUBE_PLANES\n\n");
		fclose(fp);
		exit(0);
	}

	fclose(fp);
	return 1;
}

void mem_alloc_error(void *block) {
	if (block == NULL) {
		printf("\nError allocating memory!\n\n");
		exit(0);
	}
}

